import sys
import warnings
from importlib.metadata import version, PackageNotFoundError

from .base import TasData
from .tas.taipan import Taipan
from .tas.sika import Sika
from .tas.bt7 import BT7
from .tas.bt4 import BT4

try:
    from .gui.TASDataBrowser import main
except ImportError:
    warnings.warn('QtPy not found, cannot run Resolution GUI')

try:
    __version__ = version("tasvisan")   # matches name in pyproject.toml
except PackageNotFoundError:
    __version__ = "unknown"


__all__ = ['TasData', 'Taipan', 'Sika', 'BT7', 'BT4']